import time
import json
from django.conf import settings
from django.urls import resolve
from django.utils import timezone

from mongo_drf_endpoint_logger.utils import (
    get_compiled_headers,
    check_private_data,
    get_request_ip
)

from mongo_drf_endpoint_logger.threads import LOGGER_THREAD


class EndpointLoggerMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB = False
        self.MONGO_DRF_ENDPOINT_LOGGER_PATH_TYPE = 'ABSOLUTE_URI'
        self.MONGO_DRF_ENDPOINT_LOGGER_SKIP_URL_NAME = []
        self.MONGO_DRF_ENDPOINT_LOGGER_SKIP_NAMESPACE = []
        self.MONGO_DRF_ENDPOINT_LOGGER_METHODS = []
        self.MONGO_DRF_ENDPOINT_LOGGER_STATUS_CODES = []

        check_key_type_in_settings = {
            'MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB': [bool],
            'MONGO_DRF_ENDPOINT_LOGGER_PATH_TYPE': [str],
            'MONGO_DRF_ENDPOINT_LOGGER_SKIP_URL_NAME': [list, tuple],
            'MONGO_DRF_ENDPOINT_LOGGER_SKIP_NAMESPACE': [list, tuple],
            'MONGO_DRF_ENDPOINT_LOGGER_METHODS': [list, tuple],
            'MONGO_DRF_ENDPOINT_LOGGER_STATUS_CODES': [list, tuple],
        }

        # Check settings attr with value type
        for key, key_type in check_key_type_in_settings.items():
            if hasattr(settings, key):
                value_attr = getattr(settings, key)
                if type(value_attr) in key_type:
                    setattr(self, key, value_attr)

        # TODO: i must add signal for other DB system.

    def __call__(self, request):
        if self.MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB:
            request_url_name = resolve(request.path).url_name
            request_namespace = resolve(request.path).namespace
            if (
                    request_namespace == 'admin'
                    or request_namespace in self.MONGO_DRF_ENDPOINT_LOGGER_SKIP_NAMESPACE
                    or request_url_name in self.MONGO_DRF_ENDPOINT_LOGGER_SKIP_URL_NAME
            ):
                return self.get_response(request)
            request_start_time = time.time()

            try:
                request_data = json.loads(request.body) if request.body else ''
            except ValueError as e:
                request_data = ''

            response = self.get_response(request)

            if (
                    self.MONGO_DRF_ENDPOINT_LOGGER_STATUS_CODES
                    and response.status_code not in self.MONGO_DRF_ENDPOINT_LOGGER_STATUS_CODES
            ):
                return response

            request_method = request.method
            if (
                    len(self.MONGO_DRF_ENDPOINT_LOGGER_METHODS) > 0
                    and request_method not in self.MONGO_DRF_ENDPOINT_LOGGER_METHODS
            ):
                return response

            request_headers = get_compiled_headers(request=request)
            content_type_json_values = (
                'application/json',
                'application/vnd.api+json'
            )
            # Check response content type.
            if response.get('content-type') in content_type_json_values:
                # Check response streaming value.
                if getattr(response, 'streaming', False):
                    response_body = '***Response Is Streaming***'
                else:
                    res_content = response.content.decode() if type(response.content) == bytes else response.content
                    response_body = json.loads(res_content)

                # Check request url.
                if self.MONGO_DRF_ENDPOINT_LOGGER_PATH_TYPE == 'ABSOLUTE_URI':
                    request_url = request.build_absolute_uri()
                elif self.MONGO_DRF_ENDPOINT_LOGGER_PATH_TYPE == 'FULL_PATH_URI':
                    request_url = request.get_full_path()
                elif self.MONGO_DRF_ENDPOINT_LOGGER_PATH_TYPE == 'RAW_URI':
                    request_url = request.get_raw_uri()
                else:
                    request_url = request.build_absolute_uri()

                data = dict(
                    url=request_url,
                    headers=check_private_data(request_headers),
                    body=check_private_data(request_data),
                    method=request_method,
                    ip=get_request_ip(request),
                    response=check_private_data(response_body),
                    status_code=response.status_code,
                    execution_time=time.time() - request_start_time,
                    created_date=timezone.now()
                )
                if self.MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB:
                    if LOGGER_THREAD:
                        d = data.copy()
                        d['headers'] = json.dumps(d['headers'])
                        if request_data:
                            d['body'] = json.dumps(d['body'])
                        d['response'] = json.dumps(d['response'])
                        LOGGER_THREAD.put_log_data(data=d)
            else:
                return response
        else:
            response = self.get_response(request)
        return response



import mongoengine

from mongo_drf_endpoint_logger.utils import check_logger_active


if check_logger_active():
    class ApiLog(mongoengine.Document):
        url = mongoengine.StringField(required=True, max_length=512)
        headers = mongoengine.DictField()
        body = mongoengine.DictField()
        method = mongoengine.StringField(max_length=10)
        ip = mongoengine.StringField(max_length=50)
        response = mongoengine.DictField()
        status_code = mongoengine.StringField()
        execution_time = mongoengine.StringField(max_value=8)
        created_date = mongoengine.DateTimeField()



from mongo_drf_endpoint_logger.utils import check_logger_active

LOGGER_THREAD = None

if check_logger_active():
    from mongo_drf_endpoint_logger.operations import InsertLogIntoDatabase
    import threading

    LOG_THREAD_NAME = 'insert_log_into_database'
    already_exists = False

    for t in threading.enumerate():
        if t.getName() == LOG_THREAD_NAME:
            already_exists = True

    if not already_exists:
        t = InsertLogIntoDatabase()
        t.daemon = True
        t.setName(LOG_THREAD_NAME)
        t.start()
        LOGGER_THREAD = t



import re

from django.conf import settings

# Default sensitive keys
PRIVATE_VALUE_KEYS = ['password', 'token', 'access', 'refresh']

if hasattr(settings, 'MONGO_DRF_ENDPOINT_LOGGER_EXCLUDE_KEYS'):
    value_attr = getattr(settings, 'MONGO_DRF_ENDPOINT_LOGGER_EXCLUDE_KEYS')
    if type(value_attr) is (list, tuple):
        PRIVATE_VALUE_KEYS.extend(settings.MONGO_DRF_ENDPOINT_LOGGER_EXCLUDE_KEYS)


# Compile to header items.
def get_compiled_headers(request=None):
    regex = re.compile('^HTTP_')
    return dict(
        (regex.sub('', header), value) for (header, value) in request.META.items() if header.startswith('HTTP_')
    )


def get_request_ip(request):
    try:
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.META.get('REMOTE_ADDR')
        return ip
    except ValueError:
        return ''


# Check private data
def check_private_data(data):
    if type(data) != dict:
        return data
    for key, value in data.items():
        if key in PRIVATE_VALUE_KEYS:
            data[key] = "***PRIVATE VALUE***"
        if type(value) == dict:
            data[key] = check_private_data(data[key])
        if type(value) == list:
            data[key] = [check_private_data(item) for item in data[key]]
    return data


def check_logger_active():
    logger = False
    if hasattr(settings, 'MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB'):
        logger = settings.MONGO_DRF_ENDPOINT_LOGGER_LOG_TO_DB
    return logger



import json
from datetime import datetime
from mongo_drf_endpoint_logger.models import ApiLog
from mongo_drf_endpoint_logger.threads import LOGGER_THREAD
from mongo_drf_endpoint_logger.serializers import LogInsertSerializer

"""
    url: CharField,
    headers: DictField,
    body: DictField,
    method: CharField,
    response: DictField,
    status_code: CharField: 200,
    execution_time: string: 0,
    ip: CharField: 0.0.0.0,
    created_date: DateTimeField: datetime.now()
"""


class LogData:
    def __init__(self, url, headers, body, method, response, status_code, execution_time, ip, created_date):
        self.url = url
        self.headers = headers
        self.body = body
        self.method = method
        self.ip = ip
        self.response = response
        self.status_code = status_code
        self.execution_time = execution_time
        self.created_date = created_date


def dumps_validated_dict(data):
    data['headers'] = json.dumps(data['headers'])
    data['body'] = json.dumps(data['body'])
    data['response'] = json.dumps(data['response'])
    return data


def async_insert_log(url, headers, body, method, response, status_code=200, execution_time="0", ip="0.0.0.0",
                     created_date=datetime.now()
                     ):
    log_object = LogData(url, headers, body, method, response, status_code, execution_time, ip, created_date)
    serializer = LogInsertSerializer(log_object)
    data = serializer.run_validation(serializer.data)
    if LOGGER_THREAD:
        d = data.copy()
        d = dumps_validated_dict(d)
        LOGGER_THREAD.put_log_data(data=d)


def sync_insert_log(url, headers, body, method, response, status_code=200, execution_time="0", ip="0.0.0.0",
                    created_date=datetime.now()
                    ):
    log_object = LogData(url, headers, body, method, response, status_code, execution_time, ip, created_date)
    serializer = LogInsertSerializer(log_object)
    data = serializer.run_validation(serializer.data)
    log = ApiLog(**data)
    log.save()



from rest_framework import serializers
from rest_framework.fields import empty
from mongo_drf_endpoint_logger.utils import check_private_data


class LogInsertSerializer(serializers.Serializer):
    url = serializers.CharField(max_length=99999)
    headers = serializers.DictField()
    body = serializers.DictField()
    method = serializers.CharField(min_length=3, max_length=10)
    ip = serializers.CharField(min_length=5, max_length=20)
    response = serializers.DictField()
    status_code = serializers.IntegerField()
    execution_time = serializers.CharField()
    created_date = serializers.DateTimeField()

    def run_validation(self, data=empty):
        validated_data = super().run_validation(data=data)
        validated_data['body'] = check_private_data(validated_data['body'])
        validated_data['headers'] = check_private_data(validated_data['headers'])
        validated_data['response'] = check_private_data(validated_data['response'])
        return validated_data



# MIT License
#
# Copyright (c) 2022 M. Inan Delibas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import setuptools


def get_description():
    with open("README.md", "r") as fh:
        long_description = fh.read()
    return long_description


setuptools.setup(
    name="mongo_drf_endpoint_logger",
    version="0.1.7",
    author="Inan Delibas",
    author_email="inanndelibas@gmail.com",
    description="An API Logger for your Django Rest Framework project with Mongo DB.",
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/inanpy/mongo_drf_endpoint_logger",
    packages=setuptools.find_packages(),
    install_requires=[
        "Django>=3.0.0"
        "djangorestframework>=3.0.0"
        "mongoengine>=0.24.1"
    ],
    license='MIT',
    python_requires='>=3.6',
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import time
from queue import Queue
from threading import Thread

from django.conf import settings
from django.db.utils import OperationalError

from mongo_drf_endpoint_logger.models import ApiLog


class InsertLogIntoDatabase(Thread):
    def __init__(self):
        super().__init__()

        self.MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE = 50  # Default
        if hasattr(settings, 'MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE'):
            self.MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE = settings.MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE

        if self.MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE < 1:
            raise Exception("""
            Value of MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE must be greater than 0
            """)

        self.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL = 10  # Default DB insertion interval is 10 seconds.
        if hasattr(settings, 'DRF_LOGGER_INTERVAL'):
            self.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL = settings.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL

            if self.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL < 1:
                raise Exception("""
                Value of MONGO_DRF_ENDPOINT_LOGGER_INTERVAL must be greater than 0
                """)
        self._queue = Queue(maxsize=self.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL)

    def put_log_data(self, data):
        self._queue.put(ApiLog(**data))
        if self._queue.qsize() >= self.MONGO_DRF_ENDPOINT_LOGGER_QUEUE_MAX_SIZE:
            self._start_bulk_insertion()

    @staticmethod
    def _insert_into_data_base(bulk_item):
        try:
            # Mongo Bulk Insert
            ApiLog.objects.insert(bulk_item, load_bulk=False)
        except OperationalError:
            raise Exception("""
            ApiLog model does not exists.
            """)
        except Exception as e:
            print('Mongo DRF Endpoint Logger Exception:', e)

    def _start_bulk_insertion(self):
        bulk_item = []
        while not self._queue.empty():
            bulk_item.append(self._queue.get())
        if bulk_item:
            self._insert_into_data_base(bulk_item)

    def start_queue_process(self):
        while True:
            time.sleep(self.MONGO_DRF_ENDPOINT_LOGGER_INTERVAL)
            self._start_bulk_insertion()

    def run(self) -> None:
        self.start_queue_process()
